/*

  Compiler settings for MSVC 4.2b:

  (suggested)

  preprocessory defs:  NDEBUG ( not  _DEBUG )
  calling convention: fastcall
  processor: pentium
 
   optimizations:
      Assume No Aliasing
      Global Optimizations 
      Generate Intrinsic Functions
      Favor Fast Code 
      Frame-Pointer Omission 
      Full Optimization

  In-line Function Expansion
    Only __inline

  not using precompiled headers
  not incrementally linked or compiled

*/

#include "first.h"
#include "raster_util.h"
#include "raster.h"
#include <assert.h>
#include "rast_i.h"



using namespace raster;



 









double ut_shade_magic;

















void __fastcall raster::draw_poly(const poly_alt * p)
{
  // get first 3 verts
  const vert_2d * v0 = p->verts;
  const vert_2d * v1 = v0->next;
  const vert_2d * v2 = v1->next;

  float x0 = v0->x;
  float y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  
  if( area <= min_tri_area  )
  {
     return;
  } 
  
  // put into double
  INIT_FP

  //
  // trace out edges
  //
 
  reset_spans();
  const vert_2d * e0 = v0;
  do
   {
    const vert_2d * e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  

  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP
   return;
  }
  //
  // do a cramer inverse on the 2d vertex system
  //
  float det_  =  -1.0f/area;

  float  xf_mat00 =  dy2*det_;
  float  xf_mat01 = -dx2*det_;
  float  xf_mat10 = -dy1*det_;
  float  xf_mat11 =  dx1*det_;



   LB_TO_SCREEN_SH(plb_sh,p->shades,x0,y0)
   LB_TO_SCREEN(plb_u ,p->us    ,x0,y0)
   LB_TO_SCREEN(plb_v ,p->vs    ,x0,y0)


   fill_alt[p->tmap->has_transparency ]( p->tmap  );


   RESTORE_FP
}




void __fastcall raster::draw_poly(const poly_aft * p)
{
  // get first 3 verts
  const vert_2d * v0 = p->verts;
  const vert_2d * v1 = v0->next;
  const vert_2d * v2 = v1->next;

  float  x0 = v0->x;
  float  y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  
  if( area <= min_tri_area  )
  {
     return;
  } 
  INIT_FP
  
  //
  // trace out edges
  //
  
  
  reset_spans();
  const vert_2d * e0 = v0;
  do
   {
    const vert_2d * e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  

  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP
   return;
  }
  //
  // do a cramer inverse on the 2d vertex system
  //
  float det_  =  -1.0/area;

  float  xf_mat00 =  dy2*det_;
  float  xf_mat01 = -dx2*det_;
  float  xf_mat10 = -dy1*det_;
  float  xf_mat11 =  dx1*det_;

  
  plb_sh[0] =  p->shade*linear_shade_scale + linear_shade_eps;

  LB_TO_SCREEN(plb_u ,p->us    ,x0,y0)
  LB_TO_SCREEN(plb_v ,p->vs    ,x0,y0)


  fill_aft[p->tmap->has_transparency]( p->tmap  );


   RESTORE_FP
}


// this is actually an intrinsic function (fsqrt) on an x86
extern "C"  double  __cdecl sqrt(double);

void __fastcall raster::draw_poly(const poly_outline * p)
{

   // get first 3 verts
  const vert_2d * v0 = p->verts;
  const vert_2d * v1 = v0->next;
  const vert_2d * v2 = v1->next;

  float  x0 = v0->x;
  float  y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  if( area <= min_tri_area  )
  {
     return;
  } 



 //
 // save current trace_edge func 
 // lines go beyond poly boundarys
 //
 trace_edge_func_t * cur_trace_edge_func =  trace_edge_func;
 trace_edge_func = trace_edge_clipped;

 //
 // create edge drawer
 //
 poly_afs out;
 out.shade = p->shade;
 out.translucent =0;
 out.surf = p->surf;

 vert_2d ov0, ov1, ov2, ov3;
 out.verts = &ov0;
 ov0.next = &ov1;
 ov1.next = &ov2;
 ov2.next = &ov3;
 ov3.next = &ov0;



 const vert_2d * e0 = v0;
 do
 {
    const vert_2d * e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
  
    float dx = e1->x - e0->x;
    float dy = e1->y - e0->y;

    // normalize and scale by thickness
    float d_scale = 0.5f * p->thickness/sqrt( square(dx) + square(dy) );
    dx *= d_scale;
    dy *= d_scale;

    ov0.x = e0->x + dy;
    ov0.y = e0->y - dx;

    ov1.x = e0->x - dy;
    ov1.y = e0->y + dx;

    ov3.x = e1->x + dy;
    ov3.y = e1->y - dx;

    ov2.x = e1->x - dy;
    ov2.y = e1->y + dx;



    draw_poly(&out);



    e0 = e1;

  } while( e0 != v0 );
  


  // restore
 trace_edge_func = cur_trace_edge_func;
}


void __fastcall raster::draw_poly(const poly_als * p)
{
  // get first 3 verts
  const vert_2d * v0 = p->verts;
  const vert_2d * v1 = v0->next;
  const vert_2d * v2 = v1->next;

  float  x0 = v0->x;
  float  y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  
  if( area <= min_tri_area  )
  {
     return;
  } 
  
  //
  // trace out edges
  //
  
  INIT_FP
  
  reset_spans();
  const vert_2d * e0 = v0;
  do
   {
    const vert_2d * e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  

  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP
   return;
  }
  //
  // do a cramer inverse on the 2d vertex system
  //
  float det_  =  -1.0/area;

  float  xf_mat00 =  dy2*det_;
  float  xf_mat01 = -dx2*det_;
  float  xf_mat10 = -dy1*det_;
  float  xf_mat11 =  dx1*det_;

                  
  LB_TO_SCREEN_SH(plb_sh ,p->shades    ,x0,y0)

  ut_shade_magic = fshl_hf( 52-  p->surf->shade_bits -32-8);


  fill_als( p->surf->clut  );


  RESTORE_FP
}



extern "C" void __fastcall fill_afs_16(int pixel );
extern "C" void __fastcall fill_afs_8(int pixel );

extern "C" void __fastcall fill_afsb_16(int pixel );
extern "C" void __fastcall fill_afsb_8(int pixel );


//
// called internally
//
void __fastcall fill_afs_16(const void * clut,int ishade, int is_translucent )
{
  int out_pix = ((u16 *)clut)[ ishade ];
  
  out_pix |= (out_pix<<16);
  if( !is_translucent)
  {
   fill_afs_16( out_pix  );
  }
  else
  {
   fill_afsb_16( out_pix  );
  }
}
void __fastcall fill_afs_8(const void * clut,int ishade, int is_translucent )
{
  int out_pix = ((byte *)clut)[ ishade ];
  out_pix |= (out_pix<<8)|(out_pix<<16)|(out_pix<<24);
  if( !is_translucent)
  {
   fill_afs_8( out_pix  );
  }
  else
  {
   fill_afsb_8( out_pix  );
  }
}

extern fill_flat_solid_func * fill_afs;

void __fastcall raster::draw_poly(const poly_afs * p)
{
  //
  // trace out edges
  //
  INIT_FP

  reset_spans();
  const vert_2d * v0 = p->verts;
  const vert_2d * e0 = v0;
  do
  {
    const vert_2d * e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  

  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP
   return;
  }

  int shade_range = 1 << p->surf->shade_bits;
  int ishade = d_to_int( p->shade*shade_range );
  if( ishade < 0 )
  {
   ishade=0;
  }
   // setup color
  if( ishade >= shade_range )
    ishade = shade_range -1;


 fill_afs( p->surf->clut, ishade, p->translucent );



 RESTORE_FP

}






